#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""FAL command for remoteaccess."""

import time

from iobase.xmldata import (XmlConfNode,)
from iobase.exception import (InvalidArgError, IoBaseException, NotFoundError)
#from ionas.request import Entry
from ionas.request import Entry, XmlNode, XmlConfNode

import libfal.falcmd as falcmd
from libfal.util import (RequestError,)
from fal_config import (SHARES_CONF_FILE, USERS_CONF_FILE, GROUPS_CONF_FILE,)
from pal_info import (PAL_CONFFILE_PATH, PAL_SYSINFO_PATH,
                        PAL_REMOTEACCESS_PATH, ReqTypes, Results)
from pkgaddon.fal_raps_config import (TLIMIT_SHARE_NAME, GUEST_USER,
                                        ADMIN_USER,)
from pkgaddon.fal_remotelink3_config import (CONFIG_FILE,)


REMOTEACCESS_NAME = "remoteaccess"

USER_NAME_PATH = "user/name"

SHARELIST_NAME = "sharelist"
SHARE_XML_TMPL = """<share readonly="%s" uid="%s" gid="%s">""" + \
                    """<name>%s</name><acl enable="%s"/></share>"""

SHARE_PATH = "sharelist/share"
NAME_NAME = "name"
CLIENT_ATTR = "client"
SHARE_NAME = "share"
PATH_NAME = "path"

GROUP_PATH = "grouplist/group"
USER_PATH = "userlist/user"

SHARE_XPATH_TMPL = """sharelist/share/name[text()="%s"]/.."""

CHECK_SHARE_XML_TMPL = \
        "<remoteaccess><share><path>%s</path></share></remoteaccess>"

READONLY_PATH = "permission/readonly"
DETAILPERMISSION_PATH = "permission/detailpermission"

MOUNTED_ATTR = "mounted"
READONLY_ATTR = "readonly"

IS_GUEST_OK_KEY = "is_guest_ok"
READONLY_KEY = "readonly"
READONLY_USER_LIST_KEY = "readonly_user_list"
READ_WRITE_USER_LIST_KEY = "read_write_user_list"
READONLY_GROUP_LIST_KEY = "readonly_group_list"
READ_WRITE_GROUP_LIST_KEY = "read_write_group_list"

PERMISSION_KEY_XPATH_PAIR_LIST = (
        (READONLY_USER_LIST_KEY, "readonlymember/userlist/user"),
        (READ_WRITE_USER_LIST_KEY, "readwritemember/userlist/user"),
        (READONLY_GROUP_LIST_KEY, "readonlymember/grouplist/group"),
        (READ_WRITE_GROUP_LIST_KEY, "readwritemember/grouplist/group"),
    )

UID_XPATH_TMPL = """userlist/user/name[text()="%s"]/../uid"""
GID_XPATH_TMPL = """userlist/user/name[text()="%s"]/../gid"""

SYSINFO_NAME = "sysinfo"
MODEL_XPATH = "model"
SERIES_XPATH = "series"
FWVERSION_XPATH = "fwversion"

CLIENTLIMIT_NAME = "clientlimit"
CLIENTLIMIT_PATH = "clientlimitlist/" + CLIENTLIMIT_NAME
CLIENTID_NAME = "clientid"
CLIENTIDLIST_NAME = "clientidlist"
KEY_NAME = "key"
REGISTERINFO_NAME = "registerinfo"
LOGININFO_NAME = "logininfo"
TIME_NAME = "time"

LOGININFO_XML_TMPL = "<logininfo><time>%s</time></logininfo>"

TIMELIMIT_NAME = "timelimit"
TIMELIMIT_PATH = "timelimitlist/" + TIMELIMIT_NAME
EXPIRE_TIME_PATH = "expire/time"
CLIENTMAX_NAME = "clientmax"

PATHLIST_NAME = "pathlist"

GETSHARELIST_REQ = "getsharelist"
GETSERVERINFO_REQ = "getserverinfo"
REGISTER_CLIENTLIMIT_REQ = "register_clientlimit"
AUTH_TIMELIMIT_REQ = "auth_timelimit"
GET_TIMELIMIT_REQ = "get_timelimit"

GET_CLIENTLIMIT_LIST_REQ = "get_clientlimit_list"
MODIFY_CLIENTLIMIT_REQ = "modify_clientlimit"
GET_TIMELIMIT_LIST_REQ = "get_timelimit_list"
CHECK_SESSIONID_REQ = "check_sessionid"
MODIFY_TIMELIMIT_REQ = "modify_timelimit"
CHECK_SHARE_REQ = "check_share"
GET_CONNECTION_LIST_REQ = "get_connection_list"
ADD_CONNECTION_REQ = "add_connection"


class NoShareError(Exception):
    """No Share Error."""

    pass


class NoMoreClient(Exception):
    """No More Client Error."""

    pass


class RemoteAccess(falcmd.FalCmd):
    """FAL LogExt command implementation class."""

    def _is_exist_node(self, node, path):
        """Check whether path exists in specified node"""

        try:
            node.get_node(path)
        except NotFoundError:
            return False
        return True

    def _check_body_node(self, node, type=None):
        """Check if entry's body node is valid."""

        if node.get_node_name() != REMOTEACCESS_NAME:
            raise InvalidArgError(node.get_node_name())

    def _send_request(self, target, entry, path=None):
        """Send "entry" request to "target" PAL commands."""

        if path:
            response_entry = self.send_entry_with_file(entry, target, path)
        else:
            response_entry = self.send_entry(entry, target)

        if response_entry.get_result() != Results.OK:
            raise RequestError(response_entry.get_result())

        return response_entry

    def _is_registered_timelimit_key(self, key):
        """Check if key is the registered timelimit key."""

        for timelimit in self._get_timelimit_list():
            if key == timelimit.get_value(KEY_NAME):
                return True
        return False

    def _get_uid_gid(self, user, users_config):
        """Return uid and gid correspondig "user"."""

        uid = users_config.get_node(UID_XPATH_TMPL % (user,)).get_text()
        gid = users_config.get_node(GID_XPATH_TMPL % (user,)).get_text()
        return uid, gid

    def _create_share_node(self, user, share_name, is_readonly, users_config,
                           is_acl_enabled=False):
        """Create a "share" XML node and return it."""

        readonly = XmlConfNode.FALSE_TEXT
        if is_readonly:
            readonly = XmlConfNode.TRUE_TEXT

        uid, gid = self._get_uid_gid(user, users_config)
        return XmlConfNode(SHARE_XML_TMPL % (readonly, uid, gid, share_name,
                                             is_acl_enabled))

    def _get_group_list_for_user(self, user):
        """Return group list whom user is a member of."""

        group_config = self.send_entry_with_file(Entry(ReqTypes.LOAD),
                                    PAL_CONFFILE_PATH, GROUPS_CONF_FILE).get()

        group_list = []
        for group in group_config.get_node_list(GROUP_PATH):
            try:
                group_name = group.get_node(NAME_NAME).get_text()
                member_list = [found.get_value(NAME_NAME)
                                for found in group.get_node_list(USER_PATH)]
                if user in member_list:
                    group_list.append(group_name)
            except:
                # Ignore error
                pass

        return group_list

    def _get_share_permission_info(self, share):
        """Return share's permission information dictionary."""

        request_entry = Entry(CHECK_SHARE_REQ,
                            XmlConfNode(CHECK_SHARE_XML_TMPL %
                                (share.get_node(PATH_NAME).get_text(),)))
        response_share = self._send_request(PAL_REMOTEACCESS_PATH,
                                    request_entry).get().get_node(SHARE_NAME)

        if not response_share.is_set_to_true(MOUNTED_ATTR):
            raise NoShareError

        permission_info = {}
        if (share.get_node(READONLY_PATH).is_enabled() or
            response_share.is_set_to_true(READONLY_ATTR)):

            permission_info[READONLY_KEY] = True
        else:
            permission_info[READONLY_KEY] = False

        permission_info[READONLY_USER_LIST_KEY] = []
        permission_info[READ_WRITE_USER_LIST_KEY] = []
        permission_info[READONLY_GROUP_LIST_KEY] = []
        permission_info[READ_WRITE_GROUP_LIST_KEY] = []

        detailpermission = share.get_node(DETAILPERMISSION_PATH)
        if not detailpermission.is_enabled():
            permission_info[IS_GUEST_OK_KEY] = True
            return permission_info

        permission_info[IS_GUEST_OK_KEY] = False
        for (key, xpath) in PERMISSION_KEY_XPATH_PAIR_LIST:
            permission_info[key] = [found.get_value(NAME_NAME)
                        for found in detailpermission.get_node_list(xpath)]

        return permission_info

    def _is_clientid_registered(self, clientid, key):
        """Check if "clientid" is registered."""

        for clientlimit in self._get_clientlimit_list():
            if (clientid == clientlimit.get_value(CLIENTID_NAME) and
                key == clientlimit.get_value(KEY_NAME)):

                return True
        return False

    def _is_access_permitted_user(self, user, group_list, permission_info):
        """Check if access is permitted for "user"."""

        if user == ADMIN_USER:
            return True

        if permission_info[IS_GUEST_OK_KEY]:
            return True

        if (user in permission_info[READONLY_USER_LIST_KEY] +
                permission_info[READ_WRITE_USER_LIST_KEY]):
            return True

        if user in permission_info[READONLY_USER_LIST_KEY]:
            return True

        for group in group_list:
            if (group in permission_info[READONLY_GROUP_LIST_KEY] +
                    permission_info[READ_WRITE_GROUP_LIST_KEY]):
                return True
        return False

    def _is_access_permitted(self, user, clientid, key, group_list,
            permission_info, remoteaccess_share, is_clientlimit_enabled):
        """Check "user" can access to a share corresponding to "share"
           XML data.
        """

        if not remoteaccess_share.is_enabled():
            return False

        if (is_clientlimit_enabled and
            remoteaccess_share.is_set_to_true(CLIENT_ATTR)):

            if not self._is_clientid_registered(clientid, key):
                return False

        return self._is_access_permitted_user(user, group_list,
                                                permission_info)

    def _is_readonly_access(self, user, group_list, permission_info):
        """Check if "user" can only read from a share corresponding
           to "share" xml data.
        """

        if user == ADMIN_USER:
            return False

        if permission_info[READONLY_KEY]:
            return True

        if user in permission_info[READ_WRITE_USER_LIST_KEY]:
            return False

        for group in group_list:
            if group in permission_info[READ_WRITE_GROUP_LIST_KEY]:
                return False

        if permission_info[IS_GUEST_OK_KEY]:
            return False

        return True

    def _get_sharelist(self, user, clientid, key):
        """Return sharelist XML node for "user", "clientid" and "key"."""

        users_config = self.send_entry_with_file(Entry(ReqTypes.LOAD),
                                    PAL_CONFFILE_PATH, USERS_CONF_FILE).get()
        remotelink3_config = self.send_entry_with_file(Entry(ReqTypes.LOAD),
                                    PAL_CONFFILE_PATH, CONFIG_FILE).get()
        remoteaccess_config = remotelink3_config.get_node(REMOTEACCESS_NAME)

        sharelist = XmlConfNode(SHARELIST_NAME)
        if not user and key:
            if (remoteaccess_config.get_node(TIMELIMIT_NAME).is_enabled() and
                self._is_registered_timelimit_key(key)):

                sharelist.append_child(
                        self._create_share_node(GUEST_USER, TLIMIT_SHARE_NAME,
                                                True, users_config, False))
            return sharelist

        if not user or not clientid:
            return sharelist

        shares_config = self.send_entry_with_file(Entry(ReqTypes.LOAD),
                                PAL_CONFFILE_PATH, SHARES_CONF_FILE).get()
        group_list = self._get_group_list_for_user(user)

        is_clientlimit_enabled = \
                remoteaccess_config.get_node(CLIENTLIMIT_NAME).is_enabled()
        for remoteaccess_share in \
                    remoteaccess_config.get_node_list(SHARE_PATH):
            share_name = remoteaccess_share.get_value(NAME_NAME)
            try:
                share = shares_config.get_node(SHARE_XPATH_TMPL %
                                (share_name,))
                permission_info = self._get_share_permission_info(share)

                if not self._is_access_permitted(user, clientid, key,
                                                group_list,
                                                permission_info,
                                                remoteaccess_share,
                                                is_clientlimit_enabled):
                    continue
                is_acl_enabled = False
                if self._is_exist_node(share, './permission/acl'):
                    is_acl_enabled = \
                        share.get_node('permission/acl').is_enabled()
                sharelist.append_child(
                            self._create_share_node(user, share_name,
                                    self._is_readonly_access(user, group_list,
                                                            permission_info),
                                    users_config,
                                    is_acl_enabled))
            except:
                # Ignore error
                self.exception("exception: _get_sharelist")

        return sharelist

    def _get_clientlimit_list(self):
        """Return saved clientlimit XML node list."""

        remoteaccess = self._send_request(PAL_REMOTEACCESS_PATH,
                                    Entry(GET_CLIENTLIMIT_LIST_REQ,
                                            REMOTEACCESS_NAME)).get()
        return remoteaccess.get_node_list(CLIENTLIMIT_PATH)

    def _get_match_clientlimit(self, target_clientlimit):
        """Return clientlimit entry corresponding to "target_clientlimit"."""

        for clientlimit in self._get_clientlimit_list():
            if (clientlimit.get_value(KEY_NAME) ==
                    target_clientlimit.get_value(KEY_NAME)):
                return clientlimit
        return None

    def _get_timelimit_list(self):
        """Return saved timelimit XML node list."""

        remoteaccess = self._send_request(PAL_REMOTEACCESS_PATH,
                                    Entry(GET_TIMELIMIT_LIST_REQ,
                                            REMOTEACCESS_NAME)).get()
        return remoteaccess.get_node_list(TIMELIMIT_PATH)

    def _get_match_timelimit(self, target_timelimit):
        """Return timelimit entry corresponding to "target_timelimit"."""

        for timelimit in self._get_timelimit_list():
            if (timelimit.get_value(KEY_NAME) ==
                    target_timelimit.get_value(KEY_NAME)):
                return timelimit
        return None

    def _get_updated_clientlimit(self, clientlimit, target_clientlimit):
        """Update "clientlimit"."""

        new_clientlimit = clientlimit.get_clone()
        new_clientlimit.update(target_clientlimit, override=True)
        current_time_str = str(int(time.time()))
        register_time = XmlConfNode(TIME_NAME)
        register_time.set_text(current_time_str)
        new_clientlimit.get_node(
                        REGISTERINFO_NAME).append_child(register_time)
        new_clientlimit.append_child(
                    XmlConfNode(LOGININFO_XML_TMPL % (current_time_str,)))
        return new_clientlimit

    def _is_already_registered(self, clientid_list, auth_clientid):
        """Check if "clientid" is already registered in "clientid_list"."""

        for clientid in clientid_list:
            if clientid.get_text() == auth_clientid:
                return True

        return False

    def _try_register_clientid(self, timelimit, auth_clientid):
        """Try register "clientid"."""

        timelimit_clone = timelimit.get_clone()
        try:
            clientidlist = timelimit_clone.get_node(CLIENTIDLIST_NAME)
        except:
            timelimit_clone.append_child(XmlConfNode(CLIENTIDLIST_NAME))
            clientidlist = timelimit_clone.get_node(CLIENTIDLIST_NAME)

        clientid_list = clientidlist.get_node_list(CLIENTID_NAME)
        client_max = int(timelimit_clone.get_value(CLIENTMAX_NAME, "0"))

        if self._is_already_registered(clientid_list, auth_clientid):
            return

        if client_max:
            if client_max <= len(clientid_list):
                raise NoMoreClient

        clientid = XmlConfNode(CLIENTID_NAME)
        clientid.set_text(auth_clientid)
        clientidlist.append_child(clientid)
        req_body = XmlConfNode(REMOTEACCESS_NAME)
        req_body.append_child(timelimit_clone)
        self._send_request(PAL_REMOTEACCESS_PATH,
                        Entry(MODIFY_TIMELIMIT_REQ, req_body))

    def _getsharelist_handler(self, entry):
        """Handle getsharelist request."""

        self._check_body_node(entry.get())

        user = entry.get().get_value(USER_NAME_PATH)
        clientid = entry.get().get_value(CLIENTID_NAME)
        key = entry.get().get_value(KEY_NAME)

        entry.get().append_child(self._get_sharelist(user, clientid, key))
        entry.set_result(Results.OK)
        return entry

    def _getserverinfo_handler(self, entry):
        entry.set_result(Results.OK)

        sysinfo = self.send_entry(Entry(ReqTypes.QUERY, SYSINFO_NAME),
                                    PAL_SYSINFO_PATH).get()
        product_name = sysinfo.get_node(MODEL_XPATH).get_text()
        product_series = sysinfo.get_node(SERIES_XPATH).get_text()
        firmware_version = sysinfo.get_node(FWVERSION_XPATH).get_text()

        remoteaccess_node = XmlNode("<remoteaccess/>")
        serverinfo_node = XmlNode("<serverinfo/>")
        serverinfo_node.set_attrs({"version": "1.00"})

        connectinfolist_node = XmlNode("<connectinfolist/>")
        connectinfo_node_1 = XmlNode("<connectinfo/>")
        connectinfo_node_1.set_attrs({"protocol": "http"})
        connectinfo_node_1.append_child("hostname", attrs={"fqdn":"true"}, text='rl3-.iobb.net')
        connectinfo_node_1.append_child("port", text='40580')
        connectinfo_node_2 = XmlNode("<connectinfo/>")
        connectinfo_node_2.set_attrs({"protocol": "https"})
        connectinfo_node_2.append_child("hostname", attrs={"fqdn":"true"}, text='rl3-.iobb.net')
        connectinfo_node_2.append_child("port", text='40582')

        productinfo_node = XmlNode("<productinfo/>")
        productinfo_node.append_child("name", text=product_name)
        productinfo_node.append_child("series", text=product_series)
        productinfo_node.append_child("firmversion", text=firmware_version)

        connectinfolist_node.append_child(connectinfo_node_1)
        connectinfolist_node.append_child(connectinfo_node_2)
        serverinfo_node.append_child(connectinfolist_node)
        serverinfo_node.append_child(productinfo_node)
        remoteaccess_node.append_child(serverinfo_node)
        entry.set(remoteaccess_node)
   
        return entry

    def _register_clientlimit_handler(self, entry):
        """Handle register_clientlimit request."""

        self._check_body_node(entry.get())
        try:
            target_clientlimit = entry.get().get_node(CLIENTLIMIT_NAME)
            target_clientid = \
                    target_clientlimit.get_node(CLIENTID_NAME).get_text()
            clientlimit = self._get_match_clientlimit(target_clientlimit)
            if not clientlimit:
                entry.set_result(Results.MODIFY_ERROR)
                return entry

            current_clientid = clientlimit.get_value(CLIENTID_NAME)
            if current_clientid and (current_clientid != target_clientid):
                entry.set_result(Results.MODIFY_ERROR)
                return entry

            if current_clientid == target_clientid:
                # target is already registered.
                entry.set_result(Results.OK)
                return entry

            req_body = XmlConfNode(REMOTEACCESS_NAME)
            req_body.append_child(self._get_updated_clientlimit(
                                        clientlimit, target_clientlimit))
            self._send_request(PAL_REMOTEACCESS_PATH,
                        Entry(MODIFY_CLIENTLIMIT_REQ, req_body))
            entry.set_result(Results.OK)
        except:
            self.exception("Modify clientlimit fail")
            entry.set_result(Results.MODIFY_ERROR)

        return entry

    def _auth_timelimit_handler(self, entry):
        """Handle auth_timelimit request."""

        self._check_body_node(entry.get())
        try:
            target_timelimit = entry.get().get_node(TIMELIMIT_NAME)
            timelimit = self._get_match_timelimit(target_timelimit)
            if not timelimit:
                entry.set_result(Results.AUTH_ERROR)
                return entry

            if (time.time() >=
                    int(timelimit.get_value(EXPIRE_TIME_PATH, "0"))):

                # Already Expired
                entry.set_result(Results.AUTH_ERROR)
                return entry

            self._try_register_clientid(timelimit,
                    target_timelimit.get_node(CLIENTID_NAME).get_text())
            entry.set_result(Results.OK)
        except:
            self.exception("_auth_timelimit_handler fail")
            entry.set_result(Results.AUTH_ERROR)

        return entry

    def _get_timelimit_handler(self, entry):
        """Handle get_timelimit request."""

        self._check_body_node(entry.get())
        try:
            target_timelimit = entry.get().get_node(TIMELIMIT_NAME)
            timelimit = self._get_match_timelimit(target_timelimit)
            if not timelimit:
                entry.set_result(Results.GET_ERROR)
                return entry

            target_timelimit.update(timelimit, override=True)
            entry.set_result(Results.OK)
        except:
            self.exception("_get_timelimit_handler fail")
            entry.set_result(Results.GET_ERROR)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        falcmd.FalCmd.__init__(self, option=option)

        handler_pass_to_pal = lambda entry: \
                            self.send_entry(entry, PAL_REMOTEACCESS_PATH)

        self.set_handler(GETSHARELIST_REQ, self._getsharelist_handler)
        self.set_handler(GETSERVERINFO_REQ, self._getserverinfo_handler)
        self.set_handler(REGISTER_CLIENTLIMIT_REQ,
                            self._register_clientlimit_handler)
        self.set_handler(AUTH_TIMELIMIT_REQ, self._auth_timelimit_handler)
        self.set_handler(GET_TIMELIMIT_REQ, self._get_timelimit_handler)

        self.set_handler(CHECK_SESSIONID_REQ, handler_pass_to_pal)
        self.set_handler(MODIFY_TIMELIMIT_REQ, handler_pass_to_pal)
        self.set_handler(GET_CONNECTION_LIST_REQ, handler_pass_to_pal)
        self.set_handler(ADD_CONNECTION_REQ, handler_pass_to_pal)


#
# Main
#

if __name__ == "__main__":
    RemoteAccess(option=falcmd.FalCmdOption()).handle_request()
