#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for remoteaccess."""

import os
import random
import time
import hashlib

from iobase.exception import (InvalidArgError, NotFoundError)
from iobase.xmldata import (XmlConfNode, XmlConf, ANY_CHILD_NAME,)

from pal_info import (ReqTypes, Results)
from libpal import (palcmd, util)
from pal_config import (VOL_INFO,)
from pkgaddon.pal_raps_config import (RAPS_XML_CONF_PATH, KEY_MAX,
                                        RAPS_SESSION_FILE_TMPL,
                                        KEY_MAX_DIGIT)


REMOTEACCESS_NAME = "remoteaccess"
TIMELIMITLIST_NAME = "timelimitlist"
CLIENTLIMITLIST_NAME = "clientlimitlist"
TIMELIMIT_NAME = "timelimit"
CLIENTLIMIT_NAME = "clientlimit"
KEY_NAME = "key"
PATHLIST_NAME = "pathlist"
REGISTERINFO_NAME = "registerinfo"
USERNAME_NAME = "username"
TIME_NAME = "time"
EXPIRE_TIME_PATH = "expire/time"
REGISTER_USERNAME_PATH = REGISTERINFO_NAME + "/" + USERNAME_NAME

SESSION_NAME = "session"
ID_NAME = "id"
USER_NAME_PATH = "user/name"

TIMELIMIT_KEY_PATH = TIMELIMIT_NAME + "/key"
CLIENTLIMIT_KEY_PATH = CLIENTLIMIT_NAME + "/key"

CONNECTIONLIST_NAME = "connectionlist"
STATUS_NAME = "status"
CONNECTION_NAME = "connection"
HASHEDID_NAME = "hashedid"
HASHEDID_PATH = "connection/hashedid"
CLIENTID_NAME = "clientid"
CLIENTID_PATH = "connection/clientid"
STATUS_PATH = "connection/status"
CONNECTION_PATH = "connectionlist/connection"
CONNECTION_XPATH_TMPL = \
        """connectionlist/connection/clientid[text()="%s"]/.."""

DELETE_INFO_LIST_MAP = {
        TIMELIMIT_NAME: [("pathlist", "*")],
    }

KEY_XML_TMPL = "<key>%s</key>"
HASHEDID_XML_TMPL = "<hashedid>%s</hashedid>"

DEFAULT_XML_CONFIG = \
        "<remoteaccess><timelimitlist/><clientlimitlist/>" + \
            "<connectionlist/></remoteaccess>"

SHARE_NAME = "share"
PATH_NAME = "path"
MOUNTED_ATTR = "mounted"
READONLY_ATTR = "readonly"

MNT_POINT_KEY = "mnt_point"

CHECK_SESSIONID_REQ = "check_sessionid"
GET_PINCODE_REQ = "get_pincode"
CHANGE_PINCODE_REQ = "change_pincode"
GET_TIMELIMIT_LIST_REQ = "get_timelimit_list"
GET_CLIENTLIMIT_LIST_REQ = "get_clientlimit_list"
GET_TIMELIMIT_REQ = "get_timelimit"
ADD_TIMELIMIT_REQ = "add_timelimit"
MODIFY_TIMELIMIT_REQ = "modify_timelimit"
DELETE_TIMELIMIT_REQ = "delete_timelimit"
ADD_CLIENTLIMIT_REQ = "add_clientlimit"
MODIFY_CLIENTLIMIT_REQ = "modify_clientlimit"
DELETE_CLIENTLIMIT_REQ = "delete_clientlimit"
CHECK_SHARE_REQ = "check_share"
GET_CONNECTION_LIST_REQ = "get_connection_list"
ADD_CONNECTION_REQ = "add_connection"
MODIFY_CONNECTION_REQ = "modify_connection"
DELETE_CONNECTION_REQ = "delete_connection"


class PALRemoteAccess(palcmd.PalCmd):
    """PAL remoteaccess command implementation class."""

    def _check_body_node(self, node, type=None):
        """Check if entry's body node is valid."""

        if node.get_node_name() != REMOTEACCESS_NAME:
            raise InvalidArgError(node.get_node_name())

    def _get_session_user(self, session_file):
        """Return user name regarding to "session_file"."""

        if not os.path.exists(session_file):
            raise NotFoundError(session_file)

        session_xml = XmlConf(session_file)
        user_name = session_xml.get_node(USER_NAME_PATH).get_text()
        if not user_name:
            raise NotFoundError("no user in " + str(session_file))
        return user_name

    def _load_config(self):
        """Load XML config data."""

        try:
            xml_config = XmlConf(RAPS_XML_CONF_PATH).get_root_node()
        except:
            xml_config = XmlConfNode(DEFAULT_XML_CONFIG)

        return xml_config

    def _save_config(self, xml_config):
        """Save XML config data."""

        XmlConf(xml_config).save(RAPS_XML_CONF_PATH)

    def _try_delete_expired_timelimit(self, timelimitlist):
        """Try deleting expired timelimit entry from "timelimitlist"."""

        is_deleted = False
        idx = 0
        for timelimit in timelimitlist.get_node_list(TIMELIMIT_NAME):
            if (time.time() >=
                    int(timelimit.get_value(EXPIRE_TIME_PATH, "0"))):

                timelimitlist.delete_child(TIMELIMIT_NAME, index=idx)
                is_deleted = True
            else:
                idx += 1

        return is_deleted

    def _get_all_key_list(self, xml_config):
        """Return all key list in "xml_config"."""

        key_list = []
        for key in (xml_config.get_node_list(
                        TIMELIMITLIST_NAME + "/" + TIMELIMIT_KEY_PATH) +
                    xml_config.get_node_list(
                        CLIENTLIMITLIST_NAME + "/" + CLIENTLIMIT_KEY_PATH)):

            key_list.append(key.get_text())
        return key_list

    def _to_26_decimal_char(self, value):
        """Transform "value" to 26 decimal charactor."""

        return chr(ord("A") + int(value))

    def _create_key(self):
        """Return key data for access control."""

        random.seed()
        key_value = random.randrange(0, KEY_MAX)
        key_str = ""
        for digit in range(KEY_MAX_DIGIT, -1, -1):
            base_value = pow(26, digit)
            key_str += self._to_26_decimal_char(key_value / base_value)
            key_value %= base_value

        return key_str

    def _try_deleting_old_setting(self, old_tag, new_tag):
        """Try deleting old setting in "old_tag"."""

        del_info_list = DELETE_INFO_LIST_MAP.get(old_tag.get_node_name(), [])
        for del_info in del_info_list:
            try:
                parent_node_in_new_tag = new_tag.get_node(del_info[0])
                if parent_node_in_new_tag.get_node_list(del_info[1]):
                    old_tag.delete_nodes(del_info[0], del_info[1])
            except:
                pass

    def _get_clientid(self, entry, xml_config):
        """Return clientid corresponding target in "entry"."""

        try:
            hashedid = entry.get().get_node(HASHEDID_PATH).get_text()

            for connection in xml_config.get_node_list(CONNECTION_PATH):
                clientid = connection.get_value(CLIENTID_NAME)
                if hashedid == hashlib.md5(clientid).hexdigest():
                    return clientid
        except:
            return entry.get().get_value(CLIENTID_PATH)

        # Something wrong if come here.
        return ""

    def _check_sessionid_handler(self, entry):
        """Handle check_sessionid request."""

        self._check_body_node(entry.get())
        try:
            session = entry.get().get_node(SESSION_NAME)
            id = session.get_node(ID_NAME).get_text()
            username = XmlConfNode(USERNAME_NAME)
            username.set_text(
                    self._get_session_user(RAPS_SESSION_FILE_TMPL % (id,)))
            session.append_child(username)
            entry.set_result(Results.OK)
        except NotFoundError:
            self.exception("Not Found user")
            entry.set_result(Results.CHECK_ERROR)
        except:
            self.exception("check_sessionid")
            entry.set_result(Results.CHECK_ERROR)
        return entry

    def _get_timelimit_list_handler(self, entry):
        """Handle get_timelimit_list request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            timelimitlist = xml_config.get_node(TIMELIMITLIST_NAME)

            if self._try_delete_expired_timelimit(timelimitlist):
                self._save_config(xml_config)

            entry.get().append_child(timelimitlist)
            entry.set_result(Results.OK)
        except:
            self.exception("get_timelimit_list")
            entry.set_result(Results.GET_ERROR)
        return entry

    def _get_clientlimit_list_handler(self, entry):
        """Handle get_clientlimit_list request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            entry.get().append_child(xml_config.get_node(CLIENTLIMITLIST_NAME))
            entry.set_result(Results.OK)
        except:
            self.exception("get_clientlimit_list")
            entry.set_result(Results.GET_ERROR)
        return entry

    def _get_timelimit_handler(self, entry):
        """Handle get_timelimit request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            target_timelimit = entry.get().get_node(TIMELIMIT_NAME)
            pathlist = None
            key = target_timelimit.get_value(KEY_NAME, "")
            username = target_timelimit.get_value(REGISTER_USERNAME_PATH, "")
            try:
                pathlist = target_timelimit.get_node(PATHLIST_NAME)
            except:
                pass

            timelimitlist = xml_config.get_node(TIMELIMITLIST_NAME)
            found = False
            for timelimit in timelimitlist.get_node_list(TIMELIMIT_NAME):
                if key and timelimit.get_value(KEY_NAME) == key:
                    found = True
                    break

                if (username and
                    timelimit.get_value(REGISTER_USERNAME_PATH) == username):

                    if not pathlist.compare(timelimit.get_node(PATHLIST_NAME),
                                            True, True):
                        found = True
                        break

            if found:
                entry.get().delete_children(ANY_CHILD_NAME)
                entry.get().append_child(timelimit)
                entry.set_result(Results.OK)
            else:
                entry.set_result(Results.GET_ERROR)
            return entry
        except:
            self.exception("get_timelimit")
            entry.set_result(Results.GET_ERROR)
        return entry

    def _add_timelimit_handler(self, entry):
        """Handle add_timelimit request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            timelimit = entry.get().get_node(TIMELIMIT_NAME)
            key = self._create_key()
            while key in self._get_all_key_list(xml_config):
                key = self._create_key()

            timelimit.append_child(XmlConfNode(KEY_XML_TMPL % key))

            registerinfo = timelimit.get_node(REGISTERINFO_NAME)
            current_time_str = str(int(time.time()))
            register_time = XmlConfNode(TIME_NAME)
            register_time.set_text(current_time_str)
            registerinfo.append_child(register_time)
            timelimit.append_child(registerinfo)

            xml_config.get_node(TIMELIMITLIST_NAME).append_child(
                                                        timelimit.get_clone())
            self._save_config(xml_config)
            entry.set_result(Results.OK)
        except:
            self.exception("add_timelimit")
            entry.set_result(Results.ADD_ERROR)
        return entry

    def _modify_timelimit_handler(self, entry):
        """Handle modify_timelimit request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            new_timelimit = entry.get().get_node(TIMELIMIT_NAME)
            key = new_timelimit.get_node(KEY_NAME).get_text()

            timelimitlist = xml_config.get_node(TIMELIMITLIST_NAME)
            for timelimit in timelimitlist.get_node_list(TIMELIMIT_NAME):
                if timelimit.get_value(KEY_NAME) == key:
                    self._try_deleting_old_setting(timelimit, new_timelimit)
                    timelimit.update(new_timelimit, override=True)

            self._save_config(xml_config)
            entry.set_result(Results.OK)
        except:
            self.exception("modify_timelimit")
            entry.set_result(Results.MODIFY_ERROR)
        return entry

    def _delete_timelimit_handler(self, entry):
        """Handle delete_timelimit request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            key = entry.get().get_node(TIMELIMIT_KEY_PATH).get_text()

            idx = 0
            timelimitlist = xml_config.get_node(TIMELIMITLIST_NAME)
            for timelimit in timelimitlist.get_node_list(TIMELIMIT_NAME):
                if timelimit.get_value(KEY_NAME) == key:
                    timelimitlist.delete_child(TIMELIMIT_NAME, index=idx)
                    break
                else:
                    idx += 1

            self._save_config(xml_config)
            entry.set_result(Results.OK)
        except:
            self.exception("delete_timelimit")
            entry.set_result(Results.DELETE_ERROR)
        return entry

    def _add_clientlimit_handler(self, entry):
        """Handle add_clientlimit request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            clientlimit = entry.get().get_node(CLIENTLIMIT_NAME)
            key = self._create_key()
            while key in self._get_all_key_list(xml_config):
                key = self._create_key()

            clientlimit.append_child(XmlConfNode(KEY_XML_TMPL % key))
            xml_config.get_node(CLIENTLIMITLIST_NAME).append_child(
                                                    clientlimit.get_clone())
            self._save_config(xml_config)
            entry.set_result(Results.OK)
        except:
            self.exception("delete_clientlimit")
            entry.set_result(Results.DELETE_ERROR)
        return entry

    def _modify_clientlimit_handler(self, entry):
        """Handle modify_clientlimit request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            new_clientlimit = entry.get().get_node(CLIENTLIMIT_NAME)
            key = new_clientlimit.get_node(KEY_NAME).get_text()

            clientlimitlist = xml_config.get_node(CLIENTLIMITLIST_NAME)
            for clientlimit in clientlimitlist.get_node_list(CLIENTLIMIT_NAME):
                if clientlimit.get_value(KEY_NAME) == key:
                    self._try_deleting_old_setting(clientlimit,
                                                    new_clientlimit)
                    clientlimit.update(new_clientlimit, override=True)

            self._save_config(xml_config)
            entry.set_result(Results.OK)
        except:
            self.exception("modify_clientlimit")
            entry.set_result(Results.MODIFY_ERROR)
        return entry

    def _delete_clientlimit_handler(self, entry):
        """Handle delete_clientlimit request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            key = entry.get().get_node(CLIENTLIMIT_KEY_PATH).get_text()

            idx = 0
            clientlimitlist = xml_config.get_node(CLIENTLIMITLIST_NAME)
            for clientlimit in clientlimitlist.get_node_list(CLIENTLIMIT_NAME):
                if clientlimit.get_value(KEY_NAME) == key:
                    clientlimitlist.delete_child(CLIENTLIMIT_NAME, index=idx)
                    break
                else:
                    idx += 1

            self._save_config(xml_config)
            entry.set_result(Results.OK)
        except:
            self.exception("delete_clientlimit")
            entry.set_result(Results.DELETE_ERROR)
        return entry

    def _check_share_handler(self, entry):
        """Handle check_share request."""

        self._check_body_node(entry.get())
        try:
            share = entry.get().get_node(SHARE_NAME)
            path = share.get_node(PATH_NAME).get_text()

            is_mounted = False
            is_readonly = False
            for volume in VOL_INFO:
                mnt_point = VOL_INFO[volume][MNT_POINT_KEY]

                if (path == mnt_point or
                    path.startswith(mnt_point.rstrip(os.sep) + os.sep)):

                    if util.is_mounted(mnt_point):
                        is_mounted = True
                        if util.is_mounted_as_read_only(mnt_point):
                            is_readonly = True
                    break

            share.set_bool_attr(MOUNTED_ATTR, is_mounted)
            share.set_bool_attr(READONLY_ATTR, is_readonly)
            entry.set_result(Results.OK)
        except:
            self.exception("Error _check_share_handler")
            entry.set_result(Results.CHECK_ERROR)
        return entry

    def _apply_handler(self, entry):
        """Handle apply request."""

        self._check_body_node(entry.get())
        entry.set_result(Results.OK)
        return entry

    def _get_connection_list_handler(self, entry):
        """Handle get_connection_list request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            try:
                connectionlist = xml_config.get_node(CONNECTIONLIST_NAME)
            except:
                xml_config.append_child(CONNECTIONLIST_NAME)
                connectionlist = xml_config.get_node(CONNECTIONLIST_NAME)

            for connection in connectionlist.get_node_list(CONNECTION_NAME):
                clientid = connection.get_value(CLIENTID_NAME)
                if clientid:
                    connection.append_child(
                            XmlConfNode(HASHEDID_XML_TMPL % (
                                        hashlib.md5(clientid).hexdigest())))

            entry.get().append_child(connectionlist)
            entry.set_result(Results.OK)
        except:
            self.exception("get_connection_list")
            entry.set_result(Results.GET_ERROR)
        return entry

    def _add_connection_handler(self, entry):
        """Handle add_connection request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            try:
                connectionlist = xml_config.get_node(CONNECTIONLIST_NAME)
            except:
                xml_config.append_child(CONNECTIONLIST_NAME)
                connectionlist = xml_config.get_node(CONNECTIONLIST_NAME)

            new_connection = entry.get().get_node(CONNECTION_NAME)
            clientid = new_connection.get_value(CLIENTID_NAME)
            try:
                connection = xml_config.get_node(
                                    CONNECTION_XPATH_TMPL % (clientid))
                connection.update(new_connection, override=True)
            except:
                connectionlist.append_child(new_connection.get_clone())

            self._save_config(xml_config)
            entry.set_result(Results.OK)
        except:
            self.exception("add_connection")
            entry.set_result(Results.ADD_ERROR)
        return entry

    def _modify_connection_handler(self, entry):
        """Handle modify_connection request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            clientid = self._get_clientid(entry, xml_config)
            status = entry.get().get_value(STATUS_PATH)

            connection = xml_config.get_node(
                                    CONNECTION_XPATH_TMPL % (clientid))
            connection.get_node(STATUS_NAME).set_text(status)
            self._save_config(xml_config)
            entry.set_result(Results.OK)
        except:
            self.exception("modify_connection")
            entry.set_result(Results.MODIFY_ERROR)
        return entry

    def _delete_connection_handler(self, entry):
        """Handle delete_connection request."""

        self._check_body_node(entry.get())
        try:
            xml_config = self._load_config()
            clientid = self._get_clientid(entry, xml_config)

            idx = 0
            connectionlist = xml_config.get_node(CONNECTIONLIST_NAME)
            for connection in connectionlist.get_node_list(CONNECTION_NAME):
                if connection.get_value(CLIENTID_NAME) == clientid:
                    connectionlist.delete_child(CONNECTION_NAME, index=idx)
                    break
                else:
                    idx += 1

            self._save_config(xml_config)
            entry.set_result(Results.OK)
        except:
            self.exception("delete_connection")
            entry.set_result(Results.DELETE_ERROR)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(CHECK_SESSIONID_REQ,
                            self._check_sessionid_handler)
        self.set_handler(GET_TIMELIMIT_LIST_REQ,
                            self._get_timelimit_list_handler)
        self.set_handler(GET_CLIENTLIMIT_LIST_REQ,
                            self._get_clientlimit_list_handler)
        self.set_handler(GET_TIMELIMIT_REQ, self._get_timelimit_handler)
        self.set_handler(ADD_TIMELIMIT_REQ, self._add_timelimit_handler)
        self.set_handler(MODIFY_TIMELIMIT_REQ, self._modify_timelimit_handler)
        self.set_handler(DELETE_TIMELIMIT_REQ, self._delete_timelimit_handler)
        self.set_handler(ADD_CLIENTLIMIT_REQ, self._add_clientlimit_handler)
        self.set_handler(MODIFY_CLIENTLIMIT_REQ,
                            self._modify_clientlimit_handler)
        self.set_handler(DELETE_CLIENTLIMIT_REQ,
                            self._delete_clientlimit_handler)
        self.set_handler(CHECK_SHARE_REQ,
                            self._check_share_handler)

        self.set_handler(ReqTypes.APPLY, self._apply_handler)

        self.set_handler(GET_CONNECTION_LIST_REQ,
                            self._get_connection_list_handler)
        self.set_handler(ADD_CONNECTION_REQ,
                            self._add_connection_handler)
        self.set_handler(MODIFY_CONNECTION_REQ,
                            self._modify_connection_handler)
        self.set_handler(DELETE_CONNECTION_REQ,
                            self._delete_connection_handler)


#
# Main
#

if __name__ == "__main__":
    PALRemoteAccess(option=palcmd.PalCmdOption()).handle_request()
