#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Event handler for network configuration apply."""

import ionas.request as request
import libevh.evhcmd as evhcmd
from fal_info import (FAL_SERVICECTRL_PATH, Results, ReqTypes)

from pkgaddon.remotelink3_config import (REMOTELINK3_SERVICE,)


EVENTINFO_NAME = "eventinfo"
VOLUME_NAME_PATH = "volume/name"
CATEGORY_ATTR = "category"
ACTION_ATTR = "action"
PREVIOUS_ATTR = "previous"

NETWORK_CATEGORY = "network"
RESTART_ACTION = "restart"

SERVICECTRL_REQ_XML_TMPL = """<servicectrl>
    <servicelist>
        <service><name>%s</name></service>
    </servicelist>
</servicectrl>
"""


class EvhRemoteLink3Service(evhcmd.EvhCmd):
    """Event handler to start or stop services for remotelink3 service."""

    def __send_request(self, target, type, body=""):
        """Send request entry to "target"."""

        entry = request.Entry(type, body)
        return self.send_entry(entry, target)

    def __service_control(self, name, type):
        """Control service."""

        self.__send_request(FAL_SERVICECTRL_PATH, type,
                        SERVICECTRL_REQ_XML_TMPL % (name,))

    def __suspend_resume_service(self):
        """Suspend Remote Link 3 service and resume it."""

        self.__service_control(REMOTELINK3_SERVICE, ReqTypes.SUSPEND)
        self.__service_control(REMOTELINK3_SERVICE, ReqTypes.RESUME)

    def __handle_network_event(self, eventinfo, entry):
        """Handle network event."""

        action = eventinfo.get_attr_value(ACTION_ATTR)
        is_previous = eventinfo.is_set_to_true(PREVIOUS_ATTR)

        if action == RESTART_ACTION:
            if not is_previous:
                self.__suspend_resume_service()

        entry.set_result(Results.OK)
        return entry

    def __event_handler(self, entry):
        """Handle "event" request."""

        try:
            eventinfo = entry.get().get_node(EVENTINFO_NAME)
            if eventinfo.get_attr_value(CATEGORY_ATTR) == NETWORK_CATEGORY:
                return self.__handle_network_event(eventinfo, entry)

        except:
            # Non target events
            pass

        entry.set_result(Results.OK)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        evhcmd.EvhCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.EVENT, self.__event_handler)


#
# Main
#

if __name__ == "__main__":
    EvhRemoteLink3Service(option=evhcmd.EvhCmdOption()).handle_request()
