#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

import cgi
import cgitb
import sys
import os
import uuid
import hashlib
import json
import time
from iobase.xmldata import (XmlConf,)
from ionas.request import (Entry, Request)
from fal_info import (FAL_EVENTLOG_PATH,)
from product_conf import (SUDO_CMD,)
from product_conf.reqinfo import (ReqTypes, Results)
from fal_msginfo import (get_category_str,)

cgitb.enable()


DEVCTRL_SESSION_DIR = "/var/tmp/devctrl/session"
MESSAGES_XML_PATH = "/mnt/data/xmlconf/messages.xml"
EVENTLOG_XML_PATH = "/var/log/eventlog.xml"

GUEST_NAME = "guest"
ADMIN_NAME = "admin"
USER_ATTR = "user"

CATEGORY_ATTR = "category"
TIMESTAMP_ATTR = "timestamp"
LEVEL_ATTR = "level"
CODE_NAME = "code"
MESSAGE_NAME = "message"
LOG_ARG_PATH = "arglist/arg"
MESSAGEINFO_XPATH_TMPL = """messageinfolist/messageinfo/code[text()="%s"]/.."""
EVENTNOTICE_PATH = "noticeinfo/eventnotice"

LOG_PATH = "loglist/log"


RESPONSE_BASE_DICT = {
        "response": "",
        "lang":  "ja",
        "result": 0,
    }

INVALID_RESPONSE_DICT = {
        "response": "",
        "lang":  "ja",
        "result": 1,
    }

TO_TYPE_MAP = {
        "0": 0,
        "1": 3,
        "2": 4,
    }

TO_TYPE_STRING_MAP = {
        "0": "情報",
        "1": "警告",
        "2": "エラー",
    }


def _request(fal_cmd, entry):
    response = Request(fal_cmd, filter_cmd=SUDO_CMD).request(entry)
    return response.get_entrylist().get_entry().get()


def check_session(session_id):

    if not session_id:
        raise Exception()

    if not os.path.exists(os.path.join(DEVCTRL_SESSION_DIR, session_id)):
        raise Exception()


def get_session_user(session_id):

    with open(os.path.join(DEVCTRL_SESSION_DIR, session_id)) as session:
        return session.read().strip()


def _get_session_time(session_id):

    return os.stat(os.path.join(DEVCTRL_SESSION_DIR, session_id)).st_mtime


def _update_session_time(session_id):

    os.utime(os.path.join(DEVCTRL_SESSION_DIR, session_id), None)


def _is_target_event(messageinfo, user_name):

    try:
        eventnotice_node = messageinfo.get_node(EVENTNOTICE_PATH)
        if not eventnotice_node.is_enabled():
            return False
        if user_name == ADMIN_NAME:
            return True

        if eventnotice_node.is_set_to_true(USER_ATTR):
            return True
    except:
        pass
    return False


def _get_message(log_node, messageinfo_node):

        arg_list = [arg_node.get_text() for arg_node in
                                log_node.get_node_list(LOG_ARG_PATH)]
        return messageinfo_node.get_value(MESSAGE_NAME) % tuple(arg_list)


def _get_event_dict(log_node, messages):

        event_dict = {
                "type": 0,
                "type_string": "",
                "timestamp": 0,
                "category": 0,
                "category_string": "",
                "event": 0.,
                "event_string": "",
            }

        code = log_node.get_node(CODE_NAME).get_text()
        category = log_node.get_attr_value(CATEGORY_ATTR)
        messageinfo_node = messages.get_node(MESSAGEINFO_XPATH_TMPL % (code,))

        level = log_node.get_attr_value(LEVEL_ATTR)
        event_dict["type"] = TO_TYPE_MAP[level]
        event_dict["type_string"] = TO_TYPE_STRING_MAP[level]
        event_dict["timestamp"] = int(log_node.get_attr_value(TIMESTAMP_ATTR))
        event_dict["category"] = int(category)
        event_dict["category_string"] = get_category_str(category)
        event_dict["event"] = int(code)
        event_dict["event_string"] = _get_message(log_node, messageinfo_node)
        return event_dict

def _update_event_list(session_id, start_time, max_number, index, event_list,
        is_all=True):

    if is_all:
        event_log = _request(FAL_EVENTLOG_PATH, Entry(ReqTypes.GET, "eventlog"))
    else:
        event_log = XmlConf(EVENTLOG_XML_PATH).get_root_node()

    messages = XmlConf(MESSAGES_XML_PATH).get_root_node()

    user_name = get_session_user(session_id)
    raw_event_list = []
    start_idx = 0
    for log in event_log.get_node_list(LOG_PATH):
        code = log.get_node(CODE_NAME).get_text()
        timestamp = int(log.get_attr_value(TIMESTAMP_ATTR))
        try:
            messageinfo_node = messages.get_node(
                                        MESSAGEINFO_XPATH_TMPL % (code,))
        except:
            # Normally, don't come to here. Something wrong.
            continue

        if not _is_target_event(messageinfo_node, user_name):
            continue

        if start_time >= timestamp:
            continue

        raw_event_list.append(_get_event_dict(log, messages))

    if index > start_idx:
         start_idx = index

    if max_number < 0:
        max_number = len(raw_event_list)

    event_list[:] = raw_event_list[start_idx:max_number]
    return len(event_list)


def _wait_events(session_id, timeouts):

    timeouts_time = time.time() + timeouts - 1
    max_time = time.time() + 3600
    last_check_time = _get_session_time(session_id)
    while True:
        _update_session_time(session_id)
        if os.stat(EVENTLOG_XML_PATH).st_mtime > last_check_time:
            event_list = []
            if len(event_list):
                return event_list
        last_check_time = _get_session_time(session_id)
        if last_check_time >= timeouts_time or last_check_time >= max_time:
            return []
        time.sleep(10)


# main
#

print "Content-Type: text/json\n"


try:
    request = json.loads(sys.stdin.read())
    response = dict(RESPONSE_BASE_DICT)
    response["result"] = 1

    session_id = request["session_id"]
    check_session(session_id)
    if request["request"] == "get_event_list":
        _update_session_time(session_id)
        start_time = 0
        max_number = -1
        index = -1
        if "options" in request:
            if "start_time" in request["options"]:
                start_time = request["options"]["start_time"]
            if "max_number" in request["options"]:
                start_time = request["options"]["max_number"]
            if "index" in request["index"]:
                index = request["options"]["index"]

        response["response"] = "get_event_list"
        response["session_id"] = session_id
        event_list = []

        response["event_count"] = 0
        response["event_list"] = event_list
        response["result"] = 0
    elif request["request"] == "wait_event":
        response["response"] = "wait_event"
        response["session_id"] = session_id
        response["result"] = 0
        timeouts = 3600
        if "options" in request and "timeouts" in request["options"]:
            timeouts = request["options"]["timeouts"]

        response["event_list"] = _wait_events(session_id, timeouts)

    print json.dumps(response, indent=1, encoding="utf8",
                                ensure_ascii=False).encode("utf8")
except:
    print json.dumps(INVALID_RESPONSE_DICT, indent=1, encoding="utf8",
                                ensure_ascii=False).encode("utf8")
