# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta
import os
from django.utils.translation import ugettext as _
from django import forms
from django.conf import settings
from project.view import FuncView
from project.view import StatusView
from .. fields import (
    RemoteLink3EnabledRadioField,
    RemoteLink3ListenPortNumber1TextField,
    RemoteLink3ListenPortNumber2TextField,
    RemoteLink3UPnPRadioField,
    RemoteLink3RemapRouterPortRadioField,
    RemoteLink3RouterPortNumber1TextField,
    RemoteLink3RouterPortNumber2TextField,
    RemoteLink3TerminalAccessControlEnabledRadioField,
    RemoteLink3ExpiryAccessControlEnabledRadioField,
    RemoteLink3EnabledSharedFoldersMultipleCheckboxField,
    RemoteLink3TerminalAccessControlSharedFoldersMultipleCheckboxField,
    RemoteLink3ConnectionExpiryEnabledField,
    RemoteLink3ConnectionExpiryField,
)
from project.fal.remotelink3 import FALRemoteLink3

STATUS_TABLE = {
    "SUCCESS": _(u"rl3 Status: SUCCESS"),
    "ERROR": _(u"rl3 Status: ERROR"),
}

def _apply_setting(enabled,
                   p1,
                   p2,
                   upnp,
                   remap,
                   rp1,
                   rp2,
                   cac,
                   tac,
                   fs1,
                   fs2,
                   connection_expiry_enabled,
                   connection_expiry):
    try:
        res = FALRemoteLink3().apply({
            'remoteaccess': {
                'enabled': enabled,
                'timelimit': {
                    'enabled': tac,
                },
                'clientlimit': {
                    'enabled': cac,
                },
                'sharelist': [{
                    'name': f,
                    'enabled': f in fs1,
                    'client': f in fs2,
                } for f in list(set(fs1 + fs2))],
                'refreshconnection': {
                    'enabled': connection_expiry_enabled,
                    'period': connection_expiry,
                },
            },
            'connection': {
                'port1': p1 or '',
                'port2': p2 or '',
                'upnp': {
                    'enabled': bool(int(upnp)),
                    'remap': bool(int(remap)),
                    'port1': rp1 or '',
                    'port2': rp2 or '',
                },
            },
        })
        success = res['success']
        message = res['message']
        success = success if isinstance(success, bool) else False
        message = message if isinstance(message, basestring) else u''
        return success, message
    except:
        return False, u''


class RemoteLink3SettingForm(forms.Form):

    remotelink3 = RemoteLink3EnabledRadioField()
    listen_port_1 = RemoteLink3ListenPortNumber1TextField()
    listen_port_2 = RemoteLink3ListenPortNumber2TextField()
    use_upnp = RemoteLink3UPnPRadioField()
    remap = RemoteLink3RemapRouterPortRadioField()
    router_port_1 = RemoteLink3RouterPortNumber1TextField(required=False)
    router_port_2 = RemoteLink3RouterPortNumber2TextField(required=False)
    terminal_access_control = RemoteLink3TerminalAccessControlEnabledRadioField()
    expiry_access_control = RemoteLink3ExpiryAccessControlEnabledRadioField()

    connection_expiry_enabled = RemoteLink3ConnectionExpiryEnabledField()
    connection_expiry = RemoteLink3ConnectionExpiryField(initial=7)

    def __init__(self, sharedfoldernames, *args, **kwargs):
        super(RemoteLink3SettingForm, self).__init__(*args, **kwargs)
        choices = [(sharedfoldername, sharedfoldername) for sharedfoldername in sharedfoldernames]
        self.fields['rl3_sharedfolders'] = RemoteLink3EnabledSharedFoldersMultipleCheckboxField(required=False, choices=choices)
        self.fields['tac_sharedfolders'] = RemoteLink3TerminalAccessControlSharedFoldersMultipleCheckboxField(required=False, choices=choices)

    def clean(self):
        data = self.cleaned_data
        messages = []

        if data.get('listen_port_1') == data.get('listen_port_2'):
            messages.append(_(u'Please input different values for %(value1)s and %(value2)s.') % {
                'value1': self.fields['listen_port_1'].label,
                'value2': self.fields['listen_port_2'].label,
            })

        rp1 = data.get('router_port_1')
        rp2 = data.get('router_port_2')
        if (rp1 and rp2) and (rp1 == rp2):
            messages.append(_(u'Please input different values for %(value1)s and %(value2)s.') % {
                'value1': self.fields['router_port_1'].label,
                'value2': self.fields['router_port_2'].label,
            })

        if messages:
            raise forms.ValidationError(messages)
        return data


class RL3SettingView(FuncView):

    IN_INTRODUCTION = False

    def __deco_view(view_func):
        def _view_func(self, request):
            template, dictionary = view_func(self, request)
            return {
                self.KEY_TEMPLATE: template,
                self.KEY_TEMPLATE_DIC: dictionary,
            }
        return _view_func

    def __get_initial(self):
        try:
            config = FALRemoteLink3().load()["data"]
            enabledfolders, cacfolders = [], []
            for s in config['remoteaccess']['sharelist']:
                name = s['name']
                if s['enabled']: enabledfolders.append(name)
                if s['client']: cacfolders.append(name)
            return {
                'remotelink3': config['remoteaccess']['enabled'],
                'listen_port_1': config['connection']['port1'],
                'listen_port_2': config['connection']['port2'],
                'use_upnp': str(int(config['connection']['upnp']['enabled'])),
                'remap': str(int(config['connection']['upnp']['remap'])),
                'router_port_1': config['connection']['upnp']['port1'],
                'router_port_2': config['connection']['upnp']['port2'],
                'terminal_access_control': config['remoteaccess']['clientlimit']['enabled'],
                'expiry_access_control': config['remoteaccess']['timelimit']['enabled'],
                'rl3_sharedfolders': enabledfolders,
                'tac_sharedfolders': cacfolders,
                'connection_expiry_enabled': config['remoteaccess']['refreshconnection']['enabled'],
                'connection_expiry': config['remoteaccess']['refreshconnection']['period'],
            }
        except:
            return {}

    def __get_sharedfoldernames(self):
        try:
            from project.fal.share import Share
            return sorted(Share().load().keys())
        except:
            return []

    views = ['apply']

    @__deco_view
    def view(self, request):
        sharedfoldernames = self.__get_sharedfoldernames()
        form = RemoteLink3SettingForm(sharedfoldernames, initial=self.__get_initial())
        zipped = zip(sharedfoldernames, form['rl3_sharedfolders'], form['tac_sharedfolders'])

        template = os.path.join(self.TEMPLATES, 'setting.html')
        dictionary = {
            'form': form,
            'zipped': zipped,
            'status': RL3StatusView()._status(),
        }
        return template, dictionary

    def view_apply(self, request):
        sharedfoldernames = self.__get_sharedfoldernames()
        form = RemoteLink3SettingForm(sharedfoldernames, data=request.POST)
        zipped = zip(sharedfoldernames, form['rl3_sharedfolders'], form['tac_sharedfolders'])

        with self.processing(constname='ERR_RL3_STATE_CHANGE') as proc:
            proc.break_if_state_change()
            if form.is_valid():
                data = form.cleaned_data
                result, message = _apply_setting(
                    data['remotelink3'],
                    data['listen_port_1'],
                    data['listen_port_2'],
                    data['use_upnp'],
                    data['remap'],
                    data['router_port_1'],
                    data['router_port_2'],
                    data['terminal_access_control'],
                    data['expiry_access_control'],
                    data['rl3_sharedfolders'],
                    data['tac_sharedfolders'],
                    data['connection_expiry_enabled'],
                    data['connection_expiry'],
                )
                proc.result(result)
                self.set_message(message)

                if self.IN_INTRODUCTION and result:
                    return {
                        self.KEY_REDIRECT: './menu',  # FIXME: reverse()
                    }

        template = os.path.join(self.TEMPLATES, 'setting.html')
        dictionary = {
            'form': form,
            'zipped': zipped,
            'status': RL3StatusView()._status(),
        }
        return {
            self.KEY_TEMPLATE: template,
            self.KEY_TEMPLATE_DIC: dictionary,
        }

class RL3StatusView(StatusView):
    def _status(self):
        try:
            response = FALRemoteLink3().query()
            if response['success']:
                return self._status_info(response['data']['touserver']['status'],
                                         response['data']['touserver']['code'],
                                         )
            else:
                return ""
        except:
            return ""

    def _status_info(self, status="", code=""):
        _format = "%(status)s"
        if status == "SUCCESS":
            _format = "%(status)s"
        elif status == "ERROR":
            _format = "%(status)s(%(code)s)"
        return _format % {"status": STATUS_TABLE.get(status, "-"), "code": code}
