# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta
import os
import datetime
import time
from django.utils.translation import ugettext as _
from django import forms
from django.forms.formsets import formset_factory
from django.conf import settings
from project.view import FuncView
from .. fields import (RemoteLink3ClientEntrysMultipleCheckboxField,
                       RemoteLink3TerminalOwnersEmailTextField,
                       RemoteLink3TerminalLabelTextField)
from project.fal.remotelink3 import FALRemoteLink3, get_message, logging


def _add_pincode():
    try:
        r = FALRemoteLink3().add_clientlimit()
        pincode = r['data']['pincode']
        success = r['success']
        message = r['message']
        message = message if isinstance(message, basestring) else u''
        return success, message, pincode
    except:
        return False, u'', None


def _modify_pincode(pincode, label, email):
    try:
        res = FALRemoteLink3().modify_clientlimit(pincode, label=label, emailaddress=email)
        success = res['success']
        message = res['message']
        success = success if isinstance(success, bool) else False
        message = message if isinstance(message, basestring) else u''
        return success, message
    except:
        return False, u''


def _delete_pincode(pincode, log=True):
    try:
        r = FALRemoteLink3().delete_clientlimit(pincode, log=log)
        success = r['success']
        message = r['message']
        success = success if isinstance(success, bool) else False
        message = message if isinstance(message, basestring) else u''
        return success, message
    except:
        return False, u''


def _make_qrcode(pincode):
    try:
        if pincode:
            directory = os.path.join(settings.MEDIA_ROOT, "img", "pkg", "remotelink3")
            filename = FALRemoteLink3().create_qrcode(pincode, directory)["data"]["filename"]
            return os.path.join(settings.MEDIA_URL, "img", "pkg", "remotelink3", filename)
        return None
    except:
        return None


def _send_mail(address, pincode):
    try:
        from fal_info import Results
        res = FALRemoteLink3().send_pincode(address, pincode, messages={
            Results.OK: ("MSG_RL3_SEND_CLIENTLIMIT_PINCODE", {"mail_address": address or _(u"remotelink3: client management: emailaddress is empty")}),
            "default": ("ERR_RL3_SEND_CLIENTLIMIT_PINCODE", {"mail_address": address or _(u"remotelink3: client management: emailaddress is empty")}),
        })
        success = res['success']
        message = res['message']
        success = success if isinstance(success, bool) else False
        message = message if isinstance(message, basestring) else u''
        return success, message
    except:
        return False, u''


def _get_entrys():
    entrys = []
    try:
        datetime_format = '%Y/%m/%d %H:%M:%S'  # FIXME: locale
        for e in FALRemoteLink3().get_clientlimit_list()["data"]["clientlimitlist"]:
            t = e['registerinfo']['time']
            if t: t = datetime.datetime(*time.localtime(t)[:6]).strftime(datetime_format)
            entrys.append({
                'pincode': e['pincode'],
                'clientid': e['clientid'],
                'registered': bool(e['clientid']),
                'label': e['label'],
                'username': e['registerinfo']['username'],
                'registertime': t,
            })
    except:
        pass
    return entrys


def _clientlimit_enabled():
    try:
        return FALRemoteLink3().load()["data"]['remoteaccess']['clientlimit']['enabled']
    except:
        return False


class RemoteLink3ClientManagementForm(forms.Form):

    def __init__(self, entrys, *args, **kwargs):
        super(RemoteLink3ClientManagementForm, self).__init__(*args, **kwargs)
        choices = [(e['pincode'], e['pincode']) for e in entrys]
        self.fields['entrys'] = RemoteLink3ClientEntrysMultipleCheckboxField(choices=choices)


class RemoteLink3ExtensionPINCodeForm(forms.Form):

    pincode = forms.CharField()
    label = RemoteLink3TerminalLabelTextField()
    email = RemoteLink3TerminalOwnersEmailTextField()


class RemoteLink3ClientEntryForm(forms.Form):

    pincode = forms.CharField()
    label = RemoteLink3TerminalLabelTextField(required=False)
    email = RemoteLink3TerminalOwnersEmailTextField(required=False)


class RemoteLink3TerminalManagementView(FuncView):

    def __deco_view(view_func):
        def _view_func(self, request):
            template, dictionary = view_func(self, request)
            return {
                self.KEY_TEMPLATE: template,
                self.KEY_TEMPLATE_DIC: dictionary,
            }
        return _view_func

    def __get_initial(self):
        try:
            return [
            ]
        except:
            return []

    views = ['add', 'delete', 'send']

    @__deco_view
    def view(self, request):
        entrys = _get_entrys()
        form = RemoteLink3ClientManagementForm(entrys)
        for e, cb in zip(entrys, form['entrys']):
            e['cb'] = cb

        template = os.path.join(self.TEMPLATES, 'clients.html')
        dictionary = {
            'form': form,
            'entrys': entrys,
            'clientlimit_enabled': _clientlimit_enabled(),
        }
        return template, dictionary

    @__deco_view
    def view_delete(self, request):
        entrys = _get_entrys()
        form = RemoteLink3ClientManagementForm(entrys, data=request.POST)
        for e, cb in zip(entrys, form['entrys']):
            e['cb'] = cb

        with self.processing(constname='ERR_RL3_STATE_CHANGE') as proc:
            proc.break_if_state_change()
            if form.is_valid():
                result = True
                for pincode in form.cleaned_data['entrys']:
                    _result, _message = _delete_pincode(pincode, log=False)
                    result = result and _result
                    # self.set_message(_message)
                proc.result(result)

                if proc.result():
                    name = 'MSG_RL3_DELETE_CLIENTLIMIT_ENTRY'
                else:
                    name = 'ERR_RL3_DELETE_CLIENTLIMIT_ENTRY'
                logging(name)
                self.set_message(get_message(name))

        entrys = _get_entrys()
        form = RemoteLink3ClientManagementForm(entrys)
        for e, cb in zip(entrys, form['entrys']):
            e['cb'] = cb

        template = os.path.join(self.TEMPLATES, 'clients.html')
        dictionary = {
            'form': form,
            'entrys': entrys,
            'clientlimit_enabled': _clientlimit_enabled(),
        }
        return template, dictionary

    def view_add(self, request):
        with self.processing(constname='ERR_RL3_STATE_CHANGE') as proc:
            proc.break_if_state_change()
            success, message, pincode = _add_pincode()
            proc.result(success)
            if pincode:
                return {
                    self.KEY_REDIRECT: './modify/%s' % pincode,
                }
            else:
                self.set_message(message)
                return self.view(request)


class RemoteLink3ClientEntryModifyView(FuncView):

    def __deco_view(view_func):
        def _view_func(self, request, *args, **kwargs):
            template, dictionary = view_func(self, request, *args, **kwargs)
            return {
                self.KEY_TEMPLATE: template,
                self.KEY_TEMPLATE_DIC: dictionary,
            }
        return _view_func

    def __get_initial(self, pincode):
        try:
            label, email = '', ''
            for e in FALRemoteLink3().get_clientlimit_list()['data']['clientlimitlist']:
                if e['pincode'] == pincode:
                    label = e['label']
                    email = e['emailaddress']
            return {
                'pincode': pincode,
                'label': label,
                'email': email,
            }
        except:
            return {
                'pincode': pincode,
            }

    views = ['modify', 'send', 'add']

    @__deco_view
    def view(self, request, pincode=''):
        form = RemoteLink3ClientEntryForm(initial=self.__get_initial(pincode))

        # NOTE: ここでPINコード作るとCSRFでPINコードを作られてしまうのでNG

        template = os.path.join(self.TEMPLATES, 'pincode.html')
        dictionary = {
            'form': form,
            'pincode': pincode,
            'qrcode': _make_qrcode(pincode),
            'url': '../',
        }
        return template, dictionary

    @__deco_view
    def view_modify(self, request, pincode=''):
        form = RemoteLink3ClientEntryForm(data=request.POST)

        with self.processing(constname='ERR_RL3_STATE_CHANGE') as proc:
            proc.break_if_state_change()
            if form.is_valid():
                f = form.cleaned_data
                result, message = _modify_pincode(f['pincode'], f['label'], f['email'])
                proc.result(result)
                self.set_message(message)

        template = os.path.join(self.TEMPLATES, 'pincode.html')
        dictionary = {
            'form': form,
            'pincode': pincode,
            'qrcode': _make_qrcode(pincode),
            'url': '../',
        }
        return template, dictionary

    @__deco_view
    def view_send(self, request, pincode=''):
        form = RemoteLink3ClientEntryForm(data=request.POST)

        with self.processing(constname='ERR_RL3_STATE_CHANGE') as proc:
            proc.break_if_state_change()
            if form.is_valid():
                f = form.cleaned_data
                result, message = _send_mail(f['email'], pincode)
                proc.result(result)
                self.set_message(message)

        template = os.path.join(self.TEMPLATES, 'pincode.html')
        dictionary = {
            'form': form,
            'pincode': pincode,
            'qrcode': _make_qrcode(pincode),
            'url': '../',
        }
        return template, dictionary

    def view_add(self, request, pincode=''):
        form = RemoteLink3ClientEntryForm(data=request.POST)

        with self.processing(constname='ERR_RL3_STATE_CHANGE') as proc:
            proc.break_if_state_change()
            if form.is_valid():
                success, message, new_pincode = _add_pincode()
                self.set_message(message)
                proc.result(success)
                if pincode:
                    return {self.KEY_REDIRECT: './%s' % new_pincode}
        return self.view(request, pincode)
